#!python
"""
Build and publish package.
"""

from pathlib import Path
from typing import Annotated

import typer

from admin import PROJECT_NAME, PROJECT_ROOT, SOURCE_DIR
from admin.utils import DryAnnotation, logger, run

BUILD_DIST_DIR = PROJECT_ROOT / 'dist'
VERSION_FILES = [
    PROJECT_ROOT / 'pyproject.toml',
    SOURCE_DIR / '__init__.py',
]
"""
Files that contain the package version.
This version needs to be updated with each release.
"""

app = typer.Typer(
    help=__doc__,
    no_args_is_help=True,
    add_completion=False,
    rich_markup_mode='markdown',
)


def _update_project_version(version: str):
    regex = r'''^([ _]*version[ _]*[:=] *['"])(.*)(['"].*)$'''
    for file in VERSION_FILES:
        _re_sub_file(file, regex, version)


def _get_project_version() -> str:
    import re

    pattern = re.compile('''^[ _]*version[ _]*[:=] *['"](.*)['"]''', re.MULTILINE)
    versions = {}
    for file in VERSION_FILES:
        with open(file) as f:
            text = f.read()
        match = pattern.search(text)
        if not match:
            logger.error(f'Could not find version in `{file.relative_to(PROJECT_ROOT)}`.')
            raise typer.Exit(1)
        versions[file] = match.group(1)

    if len(set(versions.values())) != 1:
        logger.error(
            'Version mismatch in files that contain versions.\n'
            + (
                '\n'.join(
                    f'{file.relative_to(PROJECT_ROOT)}: {version}'
                    for file, version in versions.items()
                )
            )
        )
        raise typer.Exit(1)

    return list(versions.values())[0]


def _get_next_version(current_version, part):
    from packaging.version import Version

    version = Version(str(current_version))

    if part == 'major':
        new_version = Version(f'{version.major + 1}.0.0')
    elif part == 'minor':
        new_version = Version(f'{version.major}.{version.minor + 1}.0')
    elif part == 'patch':
        new_version = Version(f'{version.major}.{version.minor}.{version.micro + 1}')
    else:
        raise ValueError('`part` must be "major", "minor", or "patch"')

    return new_version


def _re_sub_file(file: str | Path, regex: str, repl: str, save: bool = True) -> str:
    """
    Regex search/replace text in a file.

    :param file: File to update.
    :param regex: Regex pattern, as a string.
        The regex needs to return 3 capturing groups: text before, text to replace, text after
        (per line).
    :param repl: Text to replace with.
    :param save: Whether to save the file with the new text.
    :return: Updated text.
    """
    import re

    pattern = re.compile(regex, re.MULTILINE)
    with open(file) as f:
        text = f.read()
    new_text = pattern.sub(lambda match: f'{match.group(1)}{repl}{match.group(3)}', text)

    if save:
        with open(file, 'w') as f:
            f.write(new_text)

    return new_text


def _get_release_name_and_tag(version: str) -> tuple[str, str]:
    """
    Generate release name and tag based on the version.

    :return: Tuple with release name (ex 'v1.2.3') and tag (ex '1.2.3').
    """
    return f'v{version}', version


def _get_version_from_release_name(release_name: str) -> str:
    if not release_name.startswith('v'):
        logger.error(f'Invalid release name: {release_name}')
        raise typer.Exit(1)
    return release_name[1:]


def _get_latest_release() -> tuple[str, str, list[dict]]:
    """
    Retrieves the latest release from GitHub.

    :return: Tuple with: release name (ex 'v1.2.3'), tag (ex '1.2.3') and list of assets uploaded.
    """
    import json

    release_info_json = run(
        'gh',
        'release',
        'view',
        '--json',
        'name,tagName,assets',
        dry=False,
        capture_output=True,
    ).stdout.strip()
    release_info = json.loads(release_info_json)
    return release_info['name'], release_info['tagName'], release_info['assets']


def _get_branch():
    """Returns the current branch."""
    return run(
        'git',
        'branch',
        '--show-current',
        dry=False,
        capture_output=True,
        text=True,
    ).stdout.strip()


def _get_default_branch():
    """Returns the default branch (usually ``main``)."""
    return run(
        'gh',
        'repo',
        'view',
        '--json',
        'defaultBranchRef',
        '--jq',
        '.defaultBranchRef.name',
        dry=False,
        capture_output=True,
        text=True,
    ).stdout.strip()


def _commit(message: str, dry: bool):
    # Commit
    run('git', 'add', *VERSION_FILES, dry=dry)
    run('git', 'commit', '-m', message, dry=dry)

    # Push current branch
    branch = _get_branch()
    run('git', 'push', 'origin', branch, dry=dry)


def _create_pr(title: str, description: str, dry: bool):
    """
    Creates a PR in GitHub and merges it after checks pass.

    If checks fail, the PR will remain open and will need to be dealt with manually.
    """
    # Create PR
    default_branch = _get_default_branch()
    branch = _get_branch()
    run(
        'gh',
        'pr',
        'create',
        '--title',
        title,
        '--body',
        description,
        '--head',
        branch,
        '--base',
        default_branch,
        dry=dry,
    )

    # Merge PR after checks pass
    run('gh', 'pr', 'merge', branch, '--squash', '--auto', dry=dry)


@app.command(name='clean')
def build_clean(dry: DryAnnotation = False):
    """
    Clean build files.
    """
    from admin.utils import OS, get_os

    paths = [f'{PROJECT_NAME}.egg-info', 'dist', 'build']
    if get_os() is OS.Windows:
        # Assuming PowerShell 6.0+
        full_command = (
            'Remove-Item -Recurse -Force -ErrorAction SilentlyContinue -Path ' + ', '.join(paths)
        )
        run('pwsh', '-Command', full_command, dry=dry, check=False)
    else:
        run('rm', '-rf', *paths, dry=dry)


@app.command(name='version')
def build_version(
    version: Annotated[
        str,
        typer.Option(
            help='Version in semantic versioning format (ex 1.5.0). '
            'If `version` is set, then `bump` cannot be used.',
            show_default=False,
        ),
    ] = '',
    bump: Annotated[
        str,
        typer.Option(
            help='Portion of the version to increase, can be "major", "minor", or "patch". '
            'If `bump` is set, then `version` cannot be used.',
            show_default=False,
        ),
    ] = '',
    mode: Annotated[
        str,
        typer.Option(
            help='What do do after the files are updated:\n'
            '`nothing`: do nothing and the changes are not committed (default).\n'
            '`commit`: commit and push the changes with the message "bump version".\n'
            '`pr`: Commit, push, create and merge PR after checks pass.'
        ),
    ] = 'nothing',
    yes: Annotated[
        bool,
        typer.Option(
            help='Don\'t ask confirmation to create new branch if necessary.',
            show_default=False,
        ),
    ] = False,
    dry: DryAnnotation = False,
):
    """
    Updates the files that contain the project version to the new version.

    Optionally, commit the changes, create a PR and merge it after checks pass.
    """
    from packaging.version import Version

    mode = mode.strip().lower()
    if mode not in ['nothing', 'commit', 'pr']:
        logger.error('Invalid `mode` choice.')
        raise typer.Exit(1)

    v1 = Version(_get_project_version())
    if version and bump:
        logger.error('Either `version` or `bump` can be set, not both.')
        raise typer.Exit(1)
    if not (version or bump):
        try:
            bump = {'1': 'major', '2': 'minor', '3': 'patch'}[
                input(
                    f'Current version is `{v1}`, which portion to bump?'
                    '\n1 - Major\n2 - Minor\n3 - Patch\n> '
                )
            ]
        except KeyError:
            logger.error('Invalid choice')
            raise typer.Exit(1)

    if version:
        v2 = Version(version)
        if v2 <= v1:
            logger.error(
                f'New version `{v2}` needs to be greater than the existing version `{v1}`.'
            )
            raise typer.Exit(1)
    else:
        try:
            v2 = _get_next_version(v1, bump.strip().lower())
        except AttributeError:
            logger.error('Invalid `bump` choice.')
            raise typer.Exit(1)

    # Verify branch is not default
    branch = _get_branch()
    default_branch = _get_default_branch()
    if branch == default_branch:
        branch_ok = False
        if yes or input(
            f'Current branch `{branch}` is the default branch, create new branch? [Y/n] '
        ).strip().lower() in ['', 'y', 'yes']:
            run('git', 'checkout', '-b', f'release-{v2}', dry=dry)
            branch_ok = True
        if not branch_ok:
            logger.error(f'Cannot make changes in the default branch `{branch}`.')
            raise typer.Exit(1)

    # Update files to new version
    _update_project_version(str(v2))
    print(
        f'New version is `{v2}`. Modified files :\n'
        + '\n'.join(f'  {file.relative_to(PROJECT_ROOT)}' for file in VERSION_FILES)
    )

    # Commit/push/pr
    if mode == 'nothing':
        print('Files not committed, PR not created.')
    if mode in ['commit', 'pr']:
        print('Commit and push changes.')
        _commit(f'bump version to {v2}', dry)
    if mode == 'pr':
        pr_title = f'Release {v2}'
        print(f'Create and merge PR `{pr_title}`.')
        _create_pr(pr_title, f'Preparing for release {v2}', dry)


@app.command(name='publish')
def build_publish(
    upload: Annotated[bool, typer.Option(help='Upload to Pypi.')] = True,
    yes: Annotated[
        bool,
        typer.Option(help='Don\'t request confirmation to publish to Pypi.', show_default=False),
    ] = False,
    dry: DryAnnotation = False,
):
    """
    Build package and publish (upload) to Pypi (ie, JFrog).

    Output in ``dist`` folder.
    """
    build_clean(dry)

    # Create distribution files (source and wheel)
    run('flit', 'build', dry=dry)

    # Upload to pypi
    if upload:
        if (
            yes
            or input(f'Publishing version `{_get_project_version()}` to Pypi. Press Y to confirm. ')
            .strip()
            .lower()
            == 'y'
        ):
            run('flit', 'publish', dry=dry)
        else:
            print('Package not published to Pypi.')


if __name__ == '__main__':
    app()



#!python
"""
Python packages related tasks.
"""

from enum import StrEnum
from pathlib import Path
from typing import Annotated

import typer

from admin import PROJECT_ROOT
from admin.utils import DryAnnotation, install_package, logger, multiple_parameters, run

REQUIREMENTS_DIR = PROJECT_ROOT / 'admin' / 'requirements'

app = typer.Typer(
    help=__doc__,
    no_args_is_help=True,
    add_completion=False,
    rich_markup_mode='markdown',
)


class Requirements(StrEnum):
    """
    Requirements files.

    Order matters as most operations with multiple files need ``requirements.txt`` to be processed
    first.
    Add new requirements files here.
    """

    MAIN = 'requirements'
    DEV = 'requirements-dev'


class RequirementsType(StrEnum):

    IN = 'in'
    OUT = 'txt'


REQUIREMENTS_TASK_HELP = {
    'requirements': '`.in` file. Full name not required, just the initial name after the dash '
    f'(ex. "{Requirements.DEV.name}"). For main file use "{Requirements.MAIN.name}". '
    f'Available requirements: {", ".join(Requirements)}.'
}

RequirementsAnnotation = Annotated[
    list[str] | None,
    typer.Argument(
        help='Requirement file(s) to compile. If not set, all files are compiled.\nValues can be '
        + ', '.join([f'`{x.name.lower()}`' for x in Requirements]),
        show_default=False,
    ),
]


def _get_requirements_file(
    requirements: str | Requirements, requirements_type: str | RequirementsType
) -> Path:
    """Return the full requirements file path."""
    if isinstance(requirements, Requirements):
        reqs = requirements
    else:
        try:
            reqs = Requirements[requirements.upper()]  # noqa
        except KeyError:
            try:
                reqs = Requirements(requirements.lower())
            except ValueError:
                logger.error(f'`{requirements}` is an unknown requirements file.')
                raise typer.Exit(1)

    if isinstance(requirements_type, RequirementsType):
        reqs_type = requirements_type
    else:
        reqs_type = RequirementsType(requirements_type.lstrip('.').lower())

    return REQUIREMENTS_DIR / f'{reqs}.{reqs_type}'


def _get_requirements_files(
    requirements: list[str | Requirements] | None, requirements_type: str | RequirementsType
) -> list[Path]:
    """Get full filename+extension and sort by the order defined in ``Requirements``"""
    requirements_files = list(Requirements) if requirements is None else requirements
    return [_get_requirements_file(r, requirements_type) for r in requirements_files]


@app.command(name='compile')
def pip_compile(
    requirements: RequirementsAnnotation = None,
    clean: Annotated[
        bool,
        typer.Option(
            help=f'Delete the existing requirements `{RequirementsType.OUT.value}` files, forcing '
            f'a clean compilation.'
        ),
    ] = False,
    dry: DryAnnotation = False,
):
    """
    Compile requirements file(s).
    """
    install_package('piptools', 'pip-tools', dry=dry)

    if clean and not dry:
        for filename in _get_requirements_files(requirements, RequirementsType.OUT):
            filename.unlink(missing_ok=True)

    dry_option = ['--dry-run'] if dry else []
    for filename in _get_requirements_files(requirements, RequirementsType.IN):
        run(
            'pip-compile', '--no-header', '--no-strip-extras', *dry_option, str(filename), dry=False
        )


@app.command(name='sync')
def pip_sync(requirements: RequirementsAnnotation = None, dry: DryAnnotation = False):
    """
    Synchronize environment with requirements file.
    """
    install_package('piptools', 'pip-tools', dry=dry)
    run('pip-sync', *_get_requirements_files(requirements, RequirementsType.OUT), dry=dry)


@app.command(name='package')
def pip_package(
    requirements: RequirementsAnnotation,
    package: Annotated[
        list[str], typer.Option('--package', '-p', help='One or more packages to upgrade.')
    ],
    dry: DryAnnotation = False,
):
    """
    Upgrade one or more packages.
    """
    install_package('piptools', 'pip-tools', dry=dry)

    for filename in _get_requirements_files(requirements, RequirementsType.IN):
        run(
            'pip-compile',
            '--no-strip-extras',
            *multiple_parameters('--upgrade-package', *package),
            filename,
            dry=dry,
        )


@app.command(name='upgrade')
def pip_upgrade(requirements, dry: DryAnnotation = False):
    """
    Try to upgrade all dependencies to their latest versions.

    Equivalent to ``compile`` with ``--clean`` option.

    Use ``package`` to only upgrade individual packages,
    Ex ``pip package dev mypy flake8``.
    """
    install_package('piptools', 'pip-tools', dry=dry)

    for filename in _get_requirements_files(requirements, RequirementsType.IN):
        run(['pip-compile', '--no-strip-extras', '--upgrade', filename], dry=dry)


@app.command(name='install')
def pip_install(requirements: RequirementsAnnotation, dry: DryAnnotation = False):
    """
    Equivalent to ``pip install -r <requirements*.txt>``.

    Does not require ``pip-tools``.
    """
    requirements_files = _get_requirements_files(requirements, RequirementsType.OUT)  # type: ignore
    run('pip', 'install', *multiple_parameters('-r', *requirements_files), dry=dry)


if __name__ == '__main__':
    app()



from enum import StrEnum


class SelectionStrategy(StrEnum):
    """
    What to do when more than one item is detected.
    """

    MULTIPLE = 'multiple'
    """Show menu and allow user to select any number of items."""

    ONE = 'one'
    """Show menu and allow user to select one of the items."""

    ALL = 'all'
    """Select all the items without showing the menu."""

    FAIL = 'fail'
    """Raise ``ValueError``."""

    def multiple(self) -> bool:
        """
        Returns ``True`` if more than one item can be chosen with the current selection strategy,
        ``False`` otherwise."""
        return self in [SelectionStrategy.MULTIPLE, SelectionStrategy.ALL]


class FilteredItems(StrEnum):
    """
    What happens to items as they are filtered out.
    """

    DIM = 'dim'
    """Dim the filtered items."""

    REMOVE = 'remove'
    """Remove the filtered items from the list."""

    SHOW = 'show'
    """Show the filtered items (do nothing)."""



#!python
"""
Linting and static type checking.
"""

import typer

from admin.utils import DryAnnotation, logger, run

app = typer.Typer(
    help=__doc__,
    no_args_is_help=True,
    add_completion=False,
    rich_markup_mode='markdown',
)


@app.command(name='black')
def lint_black(path='.', dry: DryAnnotation = False):
    run('black', path, dry=dry)


@app.command(name='flake8')
def lint_flake8(path='.', dry: DryAnnotation = False):
    run('flake8', path, dry=dry)


@app.command(name='isort')
def lint_isort(path='.', dry: DryAnnotation = False):
    run('isort', path, dry=dry)


@app.command(name='mypy')
def lint_mypy(path='.', dry: DryAnnotation = False):
    run('mypy', path, dry=dry)


@app.command(name='all')
def lint_all(dry: DryAnnotation = False):
    """
    Run all linters.

    Config for each of the tools is in ``pyproject.toml``.
    """
    lint_isort(dry=dry)
    lint_black(dry=dry)
    lint_flake8(dry=dry)
    lint_mypy(dry=dry)

    logger.info('Done')


if __name__ == '__main__':
    app()



"""Select items textual app."""

from typing import Generic, Iterable

from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.widgets import Footer, Header

from .options import SelectionStrategy
from .searchable_list_widget import SearchableListWidget, SelectionItemType, SelectionType


class SearchableListApp(App, Generic[SelectionType]):
    """App that uses ``SearchableListWidget`` widget to show and select items."""

    CSS = """
    Screen {
        align: center middle;
    }

    SearchableListWidget {
        width: auto;
        height: auto;
        border: round $accent;
        padding: 1;
    }
    """

    BINDINGS = [
        Binding('escape', 'quit', 'Quit', show=False),
    ]

    def __init__(
        self,
        selections: Iterable[SelectionItemType],
        selection_strategy: SelectionStrategy,
        search_title: str | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.selections = selections
        self.selection_strategy = selection_strategy
        self.search_title = search_title

    def compose(self) -> ComposeResult:
        if not self.is_inline:
            yield Header()
        yield SearchableListWidget[SelectionType](
            *self.selections,
            title=self.search_title,
            selection_strategy=self.selection_strategy,
            selected_callback=self._action_select,
        )
        if not self.is_inline:
            yield Footer()

    def on_mount(self):
        if self.selection_strategy is SelectionStrategy.ALL:
            all_selections = self.query_one(SearchableListWidget).selection_list_widget.options
            self.exit(list(all_selections))

    def _action_select(self, selected_items: list[SelectionType]):
        """Handle the ``select`` action."""
        if selected_items:
            self.exit(selected_items)



from functools import cached_property
from typing import Callable, Generic, TypeAlias

from rich.text import TextType
from textual.app import ComposeResult
from textual.containers import Vertical
from textual.events import Key
from textual.reactive import Reactive, reactive
from textual.widget import Widget
from textual.widgets import Input, Label, SelectionList
from textual.widgets._option_list import Option  # noqa
from textual.widgets._selection_list import Selection, SelectionType  # noqa

from .options import SelectionStrategy

SelectionItemType: TypeAlias = (
    Selection[SelectionType]
    | tuple[TextType, SelectionType]
    | tuple[TextType, SelectionType, bool]
    | TextType
)


class SearchableListWidget(Widget, Generic[SelectionType]):
    """
    A selection list Textual widget with search functionality and configurable selection mode.

    Search is case-insensitive.
    """

    DEFAULT_CSS = """
    Vertical {
        height: auto;
        margin-bottom: 1;
        padding: 1;
    }

    #search_title {
        width: 1fr;
        text-align: center;
        margin-bottom: 1;
        text-style: bold;
    }
    """

    search_term: Reactive[str] = reactive('')
    """Store current search term"""

    def __init__(
        self,
        *selections: SelectionItemType,
        selection_strategy: SelectionStrategy,
        title: str | None = None,
        selected_callback: Callable[[list[SelectionType]], None] | None = None,
        **kwargs,
    ):
        """
        Initialize the selection list.

        :param selections: The content for the selection list.
            Note: While the items can be of type ``Selection`` or ``tuple``, they are internally
            converted to ``Selection`` on initialization.
        :param name: The name of the selection list.
        :param id: The ID of the selection list in the DOM.
        :param classes: The CSS classes of the selection list.
        :param disabled: Whether the selection list is disabled or not.
        """
        if len(selections) > 1:
            if selection_strategy is SelectionStrategy.FAIL:
                raise ValueError(
                    f'Multiple selections not allowed with `{SelectionStrategy.FAIL}` strategy. '
                    f'{len(selections)} found.'
                )

        super().__init__(**kwargs)

        # If a selection item is a string, make it a tuple for downstream compatibility
        _selections = [(item, item) if isinstance(item, str) else item for item in selections]

        self.selections = _selections
        self.selection_strategy = selection_strategy
        self.selected_callback: Callable[[list[SelectionType]], None] | None = selected_callback
        self.multiple = self.selection_strategy.multiple()
        self.title = title

    def compose(self) -> ComposeResult:
        with Vertical():
            if self.title:
                yield Label(self.title, id='search_title')
            yield Input(placeholder='Search...', id='search_input')
            yield SelectionList[SelectionType](
                *self.selections,  # type: ignore
                id='selection_list',
            )

    def on_mount(self):
        """Initialize the widget once mounted."""
        if len(self.selections) == 1 and self.selected_callback:
            self.selected_callback([self.selection_list_widget.get_option_at_index(0).value])

        if self.multiple:
            if self.selection_strategy is SelectionStrategy.ALL:
                self.selection_list_widget.select_all()
        else:
            # TODO: Hide buttons when selecting only one entry
            # self.selection_list_widget.styles.set_rule('.selection-list--button', {'visible': 'false'})
            # self.selection_list_widget.query_one('.selection_list--button').remove()
            pass

        self.selection_list_widget.focus()

    @cached_property
    def selection_list_widget(self) -> SelectionList[SelectionType]:
        """Access the ``SelectionList`` widget."""
        return self.query_one('#selection_list', SelectionList)

    @cached_property
    def input_widget(self) -> Input:
        return self.query_one('#search_input', Input)

    def on_key(self, event: Key):
        # Override base class `OptionList` behavior for `enter`
        if event.key == 'enter':
            event.stop()
            if self.selected_callback:
                if self.multiple:
                    # Return all selected
                    self.selected_callback(self.selection_list_widget.selected)
                else:
                    # Return the highlighted option
                    option_index = self.selection_list_widget.highlighted or 0
                    value = self.selection_list_widget.get_option_at_index(option_index).value
                    self.selected_callback([value])
        elif event.key == 'down':
            if self.input_widget.has_focus:
                self.selection_list_widget.focus()
                event.stop()
        elif event.key == 'up':
            if self.selection_list_widget.has_focus:
                selected_options = self.get_selected_options()
                if (
                    not selected_options
                    or self.selection_list_widget.highlighted == selected_options[0][0]
                ):
                    self.input_widget.focus()
                    event.stop()
            elif self.input_widget.has_focus:
                selected_options = self.get_selected_options()
                if selected_options:
                    self.selection_list_widget.action_last()
                    self.selection_list_widget.focus()
                event.stop()
        elif event.key == 'escape':
            if self.selection_list_widget.has_focus:
                self.input_widget.focus()
                event.stop()
        elif event.key in ('ctrl+space', 'left', 'right'):
            if self.multiple:
                self.selection_list_widget._toggle_highlighted_selection()  # noqa
            event.stop()
        elif event.key == 'ctrl+a':
            if self.multiple:
                self.selection_list_widget.select_all()
            event.stop()
        elif event.key == 'ctrl+d':
            if self.multiple:
                self.selection_list_widget.deselect_all()
            event.stop()
        elif event.key in ('ctrl+z', 'ctrl+backspace'):
            self.input_widget.clear()
            event.stop()
        elif event.key == 'backspace':
            text_len = len(self.input_widget.value)
            if text_len > 0:
                self.input_widget.delete(text_len - 1, text_len)
            event.stop()
        elif (len(event.key) == 1 and event.key.isalnum()) or event.key in ('space', 'backspace'):
            if self.selection_list_widget.has_focus:
                # Key without combination: add to search
                self.input_widget.post_message(event)
                event.stop()

    def on_input_changed(self, event: Input.Changed):
        """Handle changes to the search input."""
        event.stop()
        if event.input.id == 'search_input':
            self.search_term = event.value.lower()
            self._filter_items()

    def _is_match(self, option: Option) -> bool:
        """Check if the option matches the search term."""
        text = str(option.prompt).lower()
        terms = self.search_term.split()
        return all(term in text for term in terms)

    def _filter_items(self):
        """Filter items based on the search term."""
        list_widget = self.selection_list_widget
        for index, option in enumerate(list_widget.options):
            is_match = self._is_match(option)
            list_widget.get_option_at_index(index).disabled = not is_match
        list_widget.refresh()

    def select(self, selection: Selection[SelectionType] | SelectionType):
        """Override to enforce single selection mode when enabled."""
        if not self.multiple:
            self.selection_list_widget.deselect_all()

        self.selection_list_widget.select(selection)

    def get_selected_options(self) -> list[tuple[int, Option]]:
        return [
            (i, option)
            for i, option in enumerate(self.selection_list_widget.options)
            if not option.disabled
        ]



"""
python tests/manual/searchable_selection_list_select.py
"""

from enum import Enum
from textual_searchable_selectionlist.options import SelectionStrategy
from textual_searchable_selectionlist.select import select, select_enum


class SampleEnum(Enum):
    A = 'a'
    B = 'b'
    C = 'c'


def print_selected(selected: list):
    print('\n----- Selections:\n' + '\n'.join(map(str, selected)))


def test_select_items_str():
    selected = select(['John', 'Jane', 'James'], SelectionStrategy.ONE, 'This is the title')
    print_selected(selected)


def test_select_items_tuple():
    selected = select(
        [('a', 'Abc'), ('b', 'Bcd'), ('c', 'Cde')], SelectionStrategy.ONE, 'This is the title'
    )
    print_selected(selected)


def test_select_enum():
    selected = select_enum(SampleEnum, SelectionStrategy.ONE, title='This is the title')
    print_selected(selected)


def test_select_enum_exclude():
    selected = select_enum(
        SampleEnum, SelectionStrategy.ONE, exclude=[SampleEnum.B], title='This is the title'
    )
    print_selected(selected)


if __name__ == '__main__':
    test_select_enum_exclude()



from enum import Enum
from unittest.mock import MagicMock, patch

from textual.widgets._option_list import Option  # noqa

from textual_searchable_selectionlist.options import SelectionStrategy
from textual_searchable_selectionlist.select import select, select_enum


class TestSelect:
    """Tests for the ``select()`` function."""

    @patch('textual_searchable_selectionlist.select.SearchableListApp')
    def test_select_returns_empty_list_when_app_returns_none(
        self, mock_app_class: MagicMock
    ):
        """Test that select returns an empty list when the app returns None."""
        mock_app_instance = MagicMock()
        mock_app_instance.run.return_value = None
        mock_app_class.return_value = mock_app_instance

        result = select(
            selections=['item1', 'item2'],
            selection_strategy=SelectionStrategy.ONE,
        )

        assert result == []

    @patch('textual_searchable_selectionlist.select.SearchableListApp')
    def test_select_returns_list_from_app(self, mock_app_class: MagicMock):
        """Test that select returns the list from the app run method."""
        expected_result = ['selected_item']
        mock_app_instance = MagicMock()
        mock_app_instance.run.return_value = expected_result
        mock_app_class.return_value = mock_app_instance

        result = select(
            selections=['item1', 'item2'],
            selection_strategy=SelectionStrategy.ONE,
        )

        assert result == expected_result

    @patch('textual_searchable_selectionlist.select.SearchableListApp')
    def test_select_passes_correct_parameters_to_app(self, mock_app_class: MagicMock):
        """Test that select passes correct parameters to SearchableListApp."""
        selections = ['item1', 'item2', 'item3']
        strategy = SelectionStrategy.MULTIPLE
        search_title = 'Test Search'
        extra_param = 'extra_value'

        mock_app_instance = MagicMock()
        mock_app_instance.run.return_value = []
        mock_app_class.return_value = mock_app_instance

        select(
            selections=selections,
            selection_strategy=strategy,
            search_title=search_title,
            custom_param=extra_param,
        )

        mock_app_class.assert_called_once_with(
            selections,
            strategy,
            search_title=search_title,
            custom_param=extra_param,
        )

    @patch('textual_searchable_selectionlist.select.SearchableListApp')
    def test_select_sets_app_title_when_provided(self, mock_app_class: MagicMock):
        """Test that select sets the app title when app_title is provided."""
        app_title = 'Custom Title'
        mock_app_instance = MagicMock()
        mock_app_instance.run.return_value = []
        mock_app_class.return_value = mock_app_instance

        select(
            selections=['item1'],
            selection_strategy=SelectionStrategy.ONE,
            app_title=app_title,
        )

        assert mock_app_instance.title == app_title

    @patch('textual_searchable_selectionlist.select.SearchableListApp')
    def test_select_does_not_set_app_title_when_none(self, mock_app_class: MagicMock):
        """Test that select does not set app title when app_title is None."""
        mock_app_instance = MagicMock()
        mock_app_instance.run.return_value = []
        mock_app_class.return_value = mock_app_instance

        select(
            selections=['item1'],
            selection_strategy=SelectionStrategy.ONE,
            app_title=None,
        )

        # Title should not be set (no assignment should occur)
        mock_app_class.assert_called_once()

    @patch('textual_searchable_selectionlist.select.SearchableListApp')
    def test_select_passes_inline_parameter_to_run(self, mock_app_class: MagicMock):
        """Test that select passes inline parameter to app.run()."""
        mock_app_instance = MagicMock()
        mock_app_instance.run.return_value = []
        mock_app_class.return_value = mock_app_instance

        select(
            selections=['item1'],
            selection_strategy=SelectionStrategy.ONE,
            inline=True,
        )

        mock_app_instance.run.assert_called_once_with(inline=True)

    @patch('textual_searchable_selectionlist.select.SearchableListApp')
    def test_select_with_inline_false(self, mock_app_class: MagicMock):
        """Test that select passes inline=False to app.run()."""
        mock_app_instance = MagicMock()
        mock_app_instance.run.return_value = []
        mock_app_class.return_value = mock_app_instance

        select(
            selections=['item1'],
            selection_strategy=SelectionStrategy.ONE,
            inline=False,
        )

        mock_app_instance.run.assert_called_once_with(inline=False)

    @patch('textual_searchable_selectionlist.select.SearchableListApp')
    def test_select_with_different_selection_strategies(self, mock_app_class: MagicMock):
        """Test select with different selection strategies."""
        mock_app_instance = MagicMock()
        mock_app_instance.run.return_value = []
        mock_app_class.return_value = mock_app_instance

        strategies = [
            SelectionStrategy.ONE,
            SelectionStrategy.MULTIPLE,
            SelectionStrategy.ALL,
            SelectionStrategy.FAIL,
        ]

        for strategy in strategies:
            mock_app_class.reset_mock()
            select(selections=['item1'], selection_strategy=strategy)
            mock_app_class.assert_called_once()
            assert mock_app_class.call_args[0][1] == strategy

    @patch('textual_searchable_selectionlist.select.SearchableListApp')
    def test_select_with_empty_selections(self, mock_app_class: MagicMock):
        """Test select with an empty selections list."""
        mock_app_instance = MagicMock()
        mock_app_instance.run.return_value = []
        mock_app_class.return_value = mock_app_instance

        result = select(selections=[], selection_strategy=SelectionStrategy.ONE)

        assert result == []
        mock_app_class.assert_called_once_with(
            [],
            SelectionStrategy.ONE,
            search_title=None,
        )

    @patch('textual_searchable_selectionlist.select.SearchableListApp')
    def test_select_with_option_objects(self, mock_app_class: MagicMock):
        """Test that select handles Option objects in the return value."""
        option1 = Option(prompt='Item 1', id='1')
        option2 = Option(prompt='Item 2', id='2')
        expected_result = [option1, option2]

        mock_app_instance = MagicMock()
        mock_app_instance.run.return_value = expected_result
        mock_app_class.return_value = mock_app_instance

        result = select(
            selections=['item1', 'item2'],
            selection_strategy=SelectionStrategy.MULTIPLE,
        )

        assert result == expected_result


class TestSelectEnum:
    """Tests for the ``select_enum()`` function."""

    class SampleEnum(Enum):
        OPTION_A = 'value_a'
        OPTION_B = 'value_b'
        OPTION_C = 'value_c'

    class NumberEnum(Enum):
        ONE = 1
        TWO = 2
        THREE = 3

    @patch('textual_searchable_selectionlist.select.select')
    def test_select_enum_by_name_returns_enum_values(self, mock_select: MagicMock):
        """Test that select_enum returns enum values when selecting by name."""
        mock_select.return_value = ['OPTION_A', 'OPTION_B']

        result = select_enum(
            enum_type=self.SampleEnum,
            selection_strategy=SelectionStrategy.MULTIPLE,
            select_by='name',
        )

        assert result == [self.SampleEnum.OPTION_A, self.SampleEnum.OPTION_B]

    @patch('textual_searchable_selectionlist.select.select')
    def test_select_enum_by_value_returns_enum_values(self, mock_select: MagicMock):
        """Test that select_enum returns enum values when selecting by value."""
        mock_select.return_value = ['value_a', 'value_c']

        result = select_enum(
            enum_type=self.SampleEnum,
            selection_strategy=SelectionStrategy.MULTIPLE,
            select_by='value',
        )

        assert result == [self.SampleEnum.OPTION_A, self.SampleEnum.OPTION_C]

    @patch('textual_searchable_selectionlist.select.select')
    def test_select_enum_passes_correct_names_to_select(self, mock_select: MagicMock):
        """Test that select_enum passes correct names to select function."""
        mock_select.return_value = []

        select_enum(
            enum_type=self.SampleEnum,
            selection_strategy=SelectionStrategy.ONE,
            select_by='name',
        )

        call_args = mock_select.call_args
        assert set(call_args.kwargs['selections']) == {'OPTION_A', 'OPTION_B', 'OPTION_C'}

    @patch('textual_searchable_selectionlist.select.select')
    def test_select_enum_passes_correct_values_to_select(self, mock_select: MagicMock):
        """Test that select_enum passes correct values to select function."""
        mock_select.return_value = []

        select_enum(
            enum_type=self.SampleEnum,
            selection_strategy=SelectionStrategy.ONE,
            select_by='value',
        )

        call_args = mock_select.call_args
        assert set(call_args.kwargs['selections']) == {'value_a', 'value_b', 'value_c'}

    @patch('textual_searchable_selectionlist.select.select')
    def test_select_enum_excludes_specified_items(self, mock_select: MagicMock):
        """Test that select_enum excludes specified enum values."""
        mock_select.return_value = []

        select_enum(
            enum_type=self.SampleEnum,
            selection_strategy=SelectionStrategy.ONE,
            select_by='name',
            exclude=[self.SampleEnum.OPTION_B],
        )

        call_args = mock_select.call_args
        assert set(call_args.kwargs['selections']) == {'OPTION_A', 'OPTION_C'}

    @patch('textual_searchable_selectionlist.select.select')
    def test_select_enum_excludes_multiple_items(self, mock_select: MagicMock):
        """Test that select_enum excludes multiple enum values."""
        mock_select.return_value = []

        select_enum(
            enum_type=self.SampleEnum,
            selection_strategy=SelectionStrategy.ONE,
            select_by='name',
            exclude=[self.SampleEnum.OPTION_A, self.SampleEnum.OPTION_C],
        )

        call_args = mock_select.call_args
        assert call_args.kwargs['selections'] == ['OPTION_B']

    @patch('textual_searchable_selectionlist.select.select')
    def test_select_enum_with_empty_exclude(self, mock_select: MagicMock):
        """Test that select_enum works with empty exclude list."""
        mock_select.return_value = []

        select_enum(
            enum_type=self.SampleEnum,
            selection_strategy=SelectionStrategy.ONE,
            select_by='name',
            exclude=[],
        )

        call_args = mock_select.call_args
        assert set(call_args.kwargs['selections']) == {'OPTION_A', 'OPTION_B', 'OPTION_C'}

    @patch('textual_searchable_selectionlist.select.select')
    def test_select_enum_with_none_exclude(self, mock_select: MagicMock):
        """Test that select_enum works when exclude is None."""
        mock_select.return_value = []

        select_enum(
            enum_type=self.SampleEnum,
            selection_strategy=SelectionStrategy.ONE,
            select_by='name',
            exclude=None,
        )

        call_args = mock_select.call_args
        assert set(call_args.kwargs['selections']) == {'OPTION_A', 'OPTION_B', 'OPTION_C'}

    @patch('textual_searchable_selectionlist.select.select')
    def test_select_enum_uses_custom_title(self, mock_select: MagicMock):
        """Test that select_enum uses custom title when provided."""
        custom_title = 'Pick your option'
        mock_select.return_value = []

        select_enum(
            enum_type=self.SampleEnum,
            selection_strategy=SelectionStrategy.ONE,
            select_by='name',
            title=custom_title,
        )

        call_args = mock_select.call_args
        assert call_args.kwargs['search_title'] == custom_title

    @patch('textual_searchable_selectionlist.select.select')
    def test_select_enum_generates_default_title(self, mock_select: MagicMock):
        """Test that select_enum generates default title from enum name."""
        mock_select.return_value = []

        select_enum(
            enum_type=self.SampleEnum,
            selection_strategy=SelectionStrategy.ONE,
            select_by='name',
        )

        call_args = mock_select.call_args
        assert call_args.kwargs['search_title'] == 'Select from SampleEnum'

    @patch('textual_searchable_selectionlist.select.select')
    def test_select_enum_with_empty_title_generates_default(self, mock_select: MagicMock):
        """Test that empty title string triggers default title generation."""
        mock_select.return_value = []

        select_enum(
            enum_type=self.SampleEnum,
            selection_strategy=SelectionStrategy.ONE,
            select_by='name',
            title='',
        )

        call_args = mock_select.call_args
        assert call_args.kwargs['search_title'] == 'Select from SampleEnum'

    @patch('textual_searchable_selectionlist.select.select')
    def test_select_enum_passes_selection_strategy(self, mock_select: MagicMock):
        """Test that select_enum passes selection strategy correctly."""
        mock_select.return_value = []

        select_enum(
            enum_type=self.SampleEnum,
            selection_strategy=SelectionStrategy.MULTIPLE,
            select_by='name',
        )

        call_args = mock_select.call_args
        assert call_args.kwargs['selection_strategy'] == SelectionStrategy.MULTIPLE

    @patch('textual_searchable_selectionlist.select.select')
    def test_select_enum_with_default_selection_strategy(self, mock_select: MagicMock):
        """Test that select_enum uses ONE as default selection strategy."""
        mock_select.return_value = []

        select_enum(enum_type=self.SampleEnum, select_by='name')

        call_args = mock_select.call_args
        assert call_args.kwargs['selection_strategy'] == SelectionStrategy.ONE

    @patch('textual_searchable_selectionlist.select.select')
    def test_select_enum_handles_option_objects_by_name(self, mock_select: MagicMock):
        """Test that select_enum handles Option objects returned from select by name."""
        option1 = MagicMock(spec=Option)
        option1.value = 'OPTION_A'
        option2 = MagicMock(spec=Option)
        option2.value = 'OPTION_C'
        mock_select.return_value = [option1, option2]

        result = select_enum(
            enum_type=self.SampleEnum,
            selection_strategy=SelectionStrategy.MULTIPLE,
            select_by='name',
        )

        assert result == [self.SampleEnum.OPTION_A, self.SampleEnum.OPTION_C]

    @patch('textual_searchable_selectionlist.select.select')
    def test_select_enum_handles_option_objects_by_value(self, mock_select: MagicMock):
        """Test that select_enum handles Option objects returned from select by value."""
        option1 = MagicMock(spec=Option)
        option1.value = 'value_b'
        mock_select.return_value = [option1]

        result = select_enum(
            enum_type=self.SampleEnum,
            selection_strategy=SelectionStrategy.ONE,
            select_by='value',
        )

        assert result == [self.SampleEnum.OPTION_B]

    @patch('textual_searchable_selectionlist.select.select')
    def test_select_enum_returns_empty_list_when_no_selection(self, mock_select: MagicMock):
        """Test that select_enum returns empty list when no items selected."""
        mock_select.return_value = []

        result = select_enum(
            enum_type=self.SampleEnum,
            selection_strategy=SelectionStrategy.ONE,
            select_by='name',
        )

        assert result == []

    @patch('textual_searchable_selectionlist.select.select')
    def test_select_enum_with_number_enum_by_value(self, mock_select: MagicMock):
        """Test that select_enum works with numeric enum values."""
        mock_select.return_value = [1, 3]

        result = select_enum(
            enum_type=self.NumberEnum,
            selection_strategy=SelectionStrategy.MULTIPLE,
            select_by='value',
        )

        assert result == [self.NumberEnum.ONE, self.NumberEnum.THREE]

    @patch('textual_searchable_selectionlist.select.select')
    def test_select_enum_with_number_enum_by_name(self, mock_select: MagicMock):
        """Test that select_enum works with numeric enum selected by name."""
        mock_select.return_value = ['TWO']

        result = select_enum(
            enum_type=self.NumberEnum,
            selection_strategy=SelectionStrategy.ONE,
            select_by='name',
        )

        assert result == [self.NumberEnum.TWO]



from enum import Enum
from typing import Iterable, Literal, Type, TypeAlias, TypeVar

from textual.widgets._option_list import Option  # noqa

from .options import SelectionStrategy
from .searchable_list_app import SearchableListApp
from .searchable_list_widget import SelectionItemType, SelectionType

T = TypeVar('T')
T_Enum = TypeVar('T_Enum', bound=Enum)
SelectionResultType: TypeAlias = SelectionType | Option


def select(
    selections: Iterable[SelectionItemType],
    selection_strategy: SelectionStrategy,
    search_title: str | None = None,
    app_title: str | None = 'Select items',
    inline: bool = False,
    **kwargs,
) -> list[SelectionResultType]:
    """
    Select one or more items.

    See Textual's ``App.run(...)`` for more options.

    :param selections: The items to choose from.
    :param selection_strategy: What to do when more than one item is detected.
        See ``SelectionStrategy`` for more details.
    :param search_title: Title of the search.
    :param app_title: Title of the Textual app.
    :param inline: Run the app inline (under the prompt).
    """
    # if selection_strategy is SelectionStrategy.ALL:
    #     return list(selections)

    app: SearchableListApp = SearchableListApp(
        selections, selection_strategy, search_title=search_title, **kwargs
    )
    if app_title:
        app.title = app_title
    selected = app.run(inline=inline)
    return selected or []


def select_enum(
    enum_type: Type[T_Enum],
    selection_strategy: SelectionStrategy = SelectionStrategy.ONE,
    select_by: Literal['name', 'value'] = 'name',
    exclude: Iterable[T_Enum] | None = None,
    title: str = '',
) -> list[T_Enum]:
    """
    Select from an enum.

    :param enum_type: The enum type to select from.
    :param selection_strategy: What to do when more than one item is detected.
        See ``SelectionStrategy`` for more details.
    :param select_by: Whether to select by enum name or value.
    :param exclude: An iterable of enum values to exclude from selection.
    :param title: The title of the selection dialog.
    """
    selectable_items = [item for item in enum_type if item not in (exclude or [])]

    selections = (
        [item.name for item in selectable_items]
        if select_by == 'name'
        else [item.value for item in enum_type]
    )

    title = title or f'Select from {enum_type.__name__}'
    selection = select(  # type: ignore
        selections=selections,
        selection_strategy=selection_strategy,
        search_title=title,
    )
    if selection and isinstance(selection[0], Option):
        selection_values = [item.value for item in selection]  # type: ignore
    else:
        selection_values = selection

    return (
        [enum_type[x] for x in selection_values]
        if select_by == 'name'
        else [enum_type(x) for x in selection_values]
    )



import logging
import subprocess
import sys
from enum import Enum
from itertools import chain
from typing import Annotated

import typer
from rich.logging import RichHandler

from admin import PROJECT_ROOT

DryAnnotation = Annotated[
    bool,
    typer.Option(
        help='Show the command that would be run without running it.',
        show_default=False,
    ),
]


class OS(str, Enum):
    """Operating System."""

    Linux = 'linux'
    MacOS = 'mac'
    Windows = 'win'


def get_os() -> OS:
    """
    Similar to ``sys.platform`` and ``platform.system()``, but less ambiguous by returning an Enum
    instead of a string.

    Doesn't make granular distinctions of linux variants, OS versions, etc.
    """
    if sys.platform == 'darwin':
        return OS.MacOS
    if sys.platform == 'win32':
        return OS.Windows
    return OS.Linux


def run(*args, dry: bool = False, **kwargs) -> subprocess.CompletedProcess | None:
    """
    Run a CLI command synchronously (i.e., wait for the command to finish) and return the result.

    This function is a wrapper around ``subprocess.run(...)``.

    If you need access to the output, add the ``capture_output=True`` argument and do
    ``.stdout.strip()`` to get the output as a string.
    """
    logger.info(' '.join(map(str, args)))

    if dry:
        return None

    defaults = dict(
        cwd=PROJECT_ROOT,
        capture_output=False,
        text=True,
        check=True,
    )

    try:
        return subprocess.run(args, **(defaults | kwargs))  # type: ignore
    except subprocess.CalledProcessError as e:
        msg = str(e)
        if e.stdout:
            msg += f'\nSTDOUT:\n{e.stdout}'
        if e.stderr:
            msg += f'\nSTDERR:\n{e.stderr}'
        logger.error(msg)
        raise typer.Exit(1)


def run_async(*args, dry: bool = False, **kwargs) -> subprocess.Popen | None:
    """
    Starts the process and continues code execution.

    Use the following checks::

        process.poll()              # Returns None if still running, else return code
        process.wait()              # Wait for completion (blocking)
        process.terminate()         # Send SIGTERM (graceful)
        process.kill()              # Send SIGKILL (force)
        process.returncode          # Access return code after completion

    See ``subprocess.Popen(...)`` for more details.
    """
    logger.info(' '.join(map(str, args)))

    if dry:
        return None

    defaults = dict(
        cwd=PROJECT_ROOT,
    )

    try:
        return subprocess.Popen(args, **(defaults | kwargs))
    except subprocess.CalledProcessError as e:
        logger.error(e)
        raise typer.Exit(1)


def is_package_installed(package_name: str) -> bool:
    """Check if a Python package is installed."""
    import importlib.util

    return importlib.util.find_spec(package_name) is not None


def install_package(package: str, package_install: str | None = None, dry: bool = False):
    """
    Install a Python package if not already installed.

    :param package: Name of the package to check/install.
    :param package_install: Name of the package to install, if different from the name to check.
    :param dry: Show the command that would be run without running it.
    """
    if is_package_installed(package):
        logger.debug(f'Package `{package}` is already installed.')
        return

    run(sys.executable, '-m', 'pip', 'install', package_install or package, dry=dry)


def multiple_parameters(parameter: str, *options) -> list[str]:
    return list(chain.from_iterable(zip([parameter] * len(options), map(str, options))))


def get_logger(name: str | None = 'typer-invoke', level=logging.DEBUG) -> logging.Logger:
    """Set up logging configuration with Rich handler and custom formatting."""

    # Create logger
    _logger = logging.getLogger(name)
    _logger.setLevel(level)
    _logger.handlers.clear()
    handler = RichHandler(
        level=level,
        show_time=False,
        show_level=True,
        show_path=False,
        markup=True,
        rich_tracebacks=False,
    )

    # Set custom format string and add handler
    formatter = logging.Formatter(fmt='%(message)s', datefmt='[%X]')  # Time format: [HH:MM:SS]
    handler.setFormatter(formatter)
    _logger.addHandler(handler)

    # Prevent logs from being handled by root logger (avoid duplicate output)
    _logger.propagate = False

    return _logger


logger = get_logger()
